use glam::Vec3;

use crate::{
    math,
    plane::{Plane, PlaneIntersection},
};

#[derive(Default, Debug, Clone)]
pub struct MaterialId(pub String);

#[derive(Debug, Clone, Copy)]
pub struct TextureProjection {
    pub u: Plane,
    pub v: Plane,
}

impl Default for TextureProjection {
    fn default() -> Self {
        Self {
            u: Plane {
                normal: glam::vec3(1.0, 0.0, 0.0),
                offset: 0.0,
            },
            v: Plane {
                normal: glam::vec3(0.0, 1.0, 0.0),
                offset: 0.0,
            },
        }
    }
}

#[derive(Default, Debug, Clone)]
pub struct BrushPlane {
    pub plane: Plane,
    pub material: MaterialId,
    pub tex_projection: TextureProjection,
}

pub struct Brush {
    pub planes: Vec<BrushPlane>,
    raw_vertices: Vec<PlaneIntersection>,
    vertices: Vec<Vec3>,
}

impl Brush {
    pub fn new(planes: &[BrushPlane]) -> Self {
        Self {
            planes: planes.to_vec(),
            raw_vertices: vec![],
            vertices: vec![],
        }
    }

    pub fn rebuild(&mut self) {
        if self.planes.len() < 4 {
            panic!("Invalid number of planes. Minimum 4 planes required.");
        }

        // Test all permutations of brush planes for intersections
        let mut intersections = vec![];
        let mut vs: Vec<Vec3> = vec![];
        let len = self.planes.len();
        for i in 0..(len - 2) {
            for j in (i + 1)..(len - 1) {
                for k in (j + 1)..len {
                    if let Some(intersection) = PlaneIntersection::from_planes(
                        self.planes[i].plane,
                        self.planes[j].plane,
                        self.planes[k].plane,
                    ) {
                        intersections.push(intersection)
                    }
                }
            }
        }

        // Validate intersections against other brush planes
        // TODO: No need to check against planes that are part of the intersection
        for intersection in &intersections {
            let mut valid = true;
            for bplane in &self.planes {
                if !intersection.in_halfspace_mat(&bplane.plane) {
                    valid = false;
                    break;
                }
            }

            // Merge vertices
            if valid {
                let point = intersection.get_intersection_point();
                let mut duplicate = false;
                for v in &vs {
                    if v.abs_diff_eq(point, math::EPSILON) {
                        duplicate = true;
                        break;
                    }
                }
                if !duplicate {
                    vs.push(point);
                }
            }
        }

        self.vertices = vs;
        self.raw_vertices = intersections;
    }

    pub fn get_vertices(&self) -> Vec<Vec3> {
        self.vertices.clone()
    }

    pub fn get_intersections(&self) -> Vec<PlaneIntersection> {
        self.raw_vertices.clone()
    }
}
